"""Utility function for the workcell manager."""

import shutil
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

from fastapi import UploadFile
from madsci.client.event_client import default_logger
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.step_types import Step
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import (
    Workflow,
    WorkflowDefinition,
    WorkflowStatus,
)
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler


def validate_node_names(workflow: Workflow, workcell: WorkcellDefinition) -> None:
    """
    Validates that the nodes in the workflow.flowdef are in the workcell.modules
    """
    for node_name in [step.node for step in workflow.flowdef]:
        if node_name not in workcell.nodes:
            raise ValueError(f"Node {node_name} not in Workcell {workcell.name}")


def replace_locations(workcell: WorkcellDefinition, step: Step) -> None:
    """Allow the user to put location names instead of joint angle value"""


def validate_step(step: Step, state_handler: WorkcellRedisHandler) -> tuple[bool, str]:
    """Check if a step is valid based on the module's about"""
    if step.node in state_handler.get_all_nodes():
        node = state_handler.get_node(step.node)
        info = node.info
        if info is None:
            return (
                True,
                f"Node {step.node} didn't return proper about information, skipping validation",
            )
        if step.action in info.actions:
            action = info.actions[step.action]
            for action_arg in action.args.values():
                if action_arg.name not in step.args and action_arg.required:
                    return (
                        False,
                        f"Step '{step.name}': Node {step.node}'s action, '{step.action}', is missing arg '{action_arg.name}'",
                    )
                # TODO: Action arg type validation goes here
            for action_file in action.files:
                if action_file.name not in step.files and action_file.required:
                    return (
                        False,
                        f"Step '{step.name}': Node {step.node}'s action, '{step.action}', is missing file '{action_file.name}'",
                    )
            return True, f"Step '{step.name}': Validated successfully"

        return (
            False,
            f"Step '{step.name}': Node {step.node} has no action '{step.action}'",
        )
    return (
        False,
        f"Step '{step.name}': Node {step.node} is not defined in workcell",
    )


def create_workflow(
    workflow_def: WorkflowDefinition,
    workcell: WorkcellDefinition,
    state_handler: WorkcellRedisHandler,
    ownership_info: Optional[OwnershipInfo] = None,
    parameters: Optional[dict[str, Any]] = None,
) -> Workflow:
    """Pulls the workcell and builds a list of dictionary steps to be executed

    Parameters
    ----------
    workflow_def: WorkflowDefintion
        The workflow data file loaded in from the workflow yaml file

    workcell : Workcell
        The Workcell object stored in the database

    parameters: Dict
        The input to the workflow

    ownership_info: OwnershipInfo
        Information on the owner(s) of the workflow

    simulate: bool
        Whether or not to use real robots

    Returns
    -------
    steps: WorkflowRun
        a completely initialized workflow run
    """
    validate_node_names(workflow_def, workcell)
    wf_dict = workflow_def.model_dump()
    wf_dict.update(
        {
            "label": workflow_def.name,
            "ownership_info": ownership_info.model_dump(mode="json"),
            "parameter_values": parameters,
        }
    )
    wf = Workflow(**wf_dict)
    steps = []
    for step in workflow_def.flowdef:
        replace_locations(workcell, step)
        valid, validation_string = validate_step(step, state_handler=state_handler)
        default_logger.log_info(validation_string)
        if not valid:
            raise ValueError(validation_string)
        steps.append(step)

    wf.steps = steps
    wf.submitted_time = datetime.now()
    return wf


def save_workflow_files(
    working_directory: str, workflow: Workflow, files: list[UploadFile]
) -> Workflow:
    """Saves the files to the workflow run directory,
    and updates the step files to point to the new location"""

    get_workflow_inputs_directory(
        workflow_id=workflow.workflow_id, working_directory=working_directory
    ).mkdir(parents=True, exist_ok=True)
    if files:
        for file in files:
            file_path = (
                get_workflow_inputs_directory(
                    working_directory=working_directory,
                    workflow_id=workflow.workflow_id,
                )
                / file.filename
            )
            with Path.open(file_path, "wb") as f:
                f.write(file.file.read())
            for step in workflow.steps:
                for step_file_key, step_file_path in step.files.items():
                    if step_file_path == file.filename:
                        step.files[step_file_key] = str(file_path)
                        default_logger.log_info(
                            f"{step_file_key}: {file_path} ({step_file_path})"
                        )
    return workflow


def copy_workflow_files(
    working_directory: str, old_id: str, workflow: Workflow
) -> Workflow:
    """Saves the files to the workflow run directory,
    and updates the step files to point to the new location"""

    new = get_workflow_inputs_directory(
        workflow_id=workflow.workflow_id, working_directory=working_directory
    )
    old = get_workflow_inputs_directory(
        workflow_id=old_id, working_directory=working_directory
    )
    shutil.copytree(old, new)
    return workflow


def get_workflow_inputs_directory(
    workflow_id: Optional[str] = None, working_directory: Optional[str] = None
) -> Path:
    """returns a directory name for the workflows inputs"""
    return Path(working_directory) / "Workflows" / workflow_id / "Inputs"


def cancel_workflow(wf: Workflow, state_handler: WorkcellRedisHandler) -> None:
    """Cancels the workflow run"""
    wf.status = WorkflowStatus.CANCELLED
    with state_handler.wc_state_lock():
        state_handler.set_workflow(wf)
    return wf


def cancel_active_workflows(state_handler: WorkcellRedisHandler) -> None:
    """Cancels all currently running workflow runs"""
    for wf in state_handler.get_all_workflows().values():
        if wf.status in [
            WorkflowStatus.RUNNING,
            WorkflowStatus.QUEUED,
            WorkflowStatus.IN_PROGRESS,
        ]:
            cancel_workflow(wf, state_handler=state_handler)



"""the abstract class for schedulers"""

from madsci.common.types.event_types import Event
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler


def send_event(test: Event) -> None:  # TODO: remove placeholder
    """send an event to the server"""


class AbstractScheduler:
    """abstract definition of a scheduler"""

    def __init__(
        self,
        workcell_definition: WorkcellDefinition,
        state_handler: WorkcellRedisHandler,
    ) -> "AbstractScheduler":
        """sets the state handler and workcell definition"""
        self.state_handler = state_handler
        self.workcell_definition = workcell_definition
        self.running = True

    def run_iteration(self) -> None:
        """run an iteration of the scheduler"""



"""Default MADSci Workcell scheduler"""

from madsci.common.types.step_types import Step
from schedulers.scheduler import AbstractScheduler


class Scheduler(AbstractScheduler):
    """the main class that handles checking whether steps are ready to run and assigning priority"""

    def run_iteration(self) -> None:
        """run an iteration of the scheduler and set priority for which workflow to run next"""
        priority = 0
        workflows = sorted(
            self.state_handler.get_all_workflows().values(),
            key=lambda item: item.submitted_time,
        )
        for wf in workflows:
            step = wf.steps[wf.step_index]
            wf.scheduler_metadata.ready_to_run = (
                not (wf.paused)
                and wf.status in ["queued", "in_progress"]
                and self.check_step(step)
            )
            wf.scheduler_metadata.priority = priority
            priority -= 1
            self.state_handler.set_workflow_quiet(wf)

    def check_step(self, step: Step) -> bool:
        """check if a step is ready to run"""
        return self.resource_checks(step) and self.node_checks(step)

    def resource_checks(self, step: Step) -> bool:  # noqa: ARG002
        """check if the resources for the step are ready TODO: actually check"""
        return True

    def node_checks(self, step: Step) -> bool:
        """check if the step node is ready to run a step"""
        node = self.state_handler.get_node(step.node)
        return node is not None and node.status.ready



"""MADSci Workcell Manager Server."""

import json
import traceback
from collections.abc import AsyncGenerator
from datetime import datetime
from typing import Annotated, Optional, Union

from fastapi import FastAPI, Form, HTTPException, UploadFile
from madsci.common.types.action_types import ActionStatus
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import new_ulid_str
from madsci.common.types.node_types import Node, NodeDefinition
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import (
    Workflow,
    WorkflowDefinition,
    WorkflowStatus,
)
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler
from madsci.workcell_manager.workcell_engine import Engine
from madsci.workcell_manager.workcell_utils import find_node_client
from madsci.workcell_manager.workflow_utils import (
    copy_workflow_files,
    create_workflow,
    save_workflow_files,
)


async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    """start the server functionality and initialize the state handler"""
    app.state.state_handler = WorkcellRedisHandler(workcell)
    app.state.state_handler.set_workcell(workcell)
    engine = Engine(workcell, app.state.state_handler)
    engine.start_engine_thread()
    yield


app = FastAPI(lifespan=lifespan)


@app.get("/definition")
@app.get("/workcell")
def get_workcell() -> WorkcellDefinition:
    """Get the currently running workcell."""
    return app.state.state_handler.get_workcell()


@app.get("/nodes")
def get_nodes() -> dict[str, Node]:
    """Get info on the nodes in the workcell."""
    return app.state.state_handler.get_all_nodes()


@app.get("/nodes/{node_name}")
def get_node(node_name: str) -> Union[Node, str]:
    """Get information about about a specific node."""
    try:
        node = app.state.state_handler.get_node(node_name)
    except Exception:
        return "Node not found!"
    return node


@app.post("/nodes/add_node")
def add_node(
    node_name: str,
    node_url: str,
    node_description: str = "A Node",
    permanent: bool = False,
) -> Union[Node, str]:
    """Add a node to the workcells nodes"""
    if node_name in app.state.state_handler.get_all_nodes():
        return "Node name exists, node names must be unique!"
    node = Node(node_url=node_url)
    app.state.state_handler.set_node(node_name, node)
    if permanent:
        workcell.nodes[node_name] = NodeDefinition(
            node_name=node_name, node_url=node_url, node_description=node_description
        )
        workcell.to_yaml(workcell._definition_path)
    return app.state.state_handler.get_node(node_name)


# TODO add node reserve endpoint


@app.get("/admin/{command}")
def send_admin_command(command: str) -> list:
    """Send an admin command to all capable nodes."""
    responses = []
    for node in app.state.state_handler.get_all_nodes().values():
        if command in node.info.capabilities.admin_commands:
            client = find_node_client(node.node_url)
            response = client.send_admin_command(command)
            responses.append(response)
    return responses


@app.get("/admin/{command}/{node}")
def send_admin_command_to_node(command: str, node: str) -> list:
    """Send admin command to a node."""
    responses = []
    node = app.state.state_handler.get_node(node)
    if command in node.info.capabilities.admin_commands:
        client = find_node_client(node.node_url)
        response = client.send_admin_command(command)
        responses.append(response)
    return responses


@app.get("/workflows")
def get_all_workflows() -> dict[str, Workflow]:
    """get all workflows."""
    return app.state.state_handler.get_all_workflows()


@app.get("/workflows/{workflow_id}")
def get_workflow(workflow_id: str) -> Workflow:
    """Get info on a specific workflow."""
    return app.state.state_handler.get_workflow(workflow_id)


@app.get("/workflows/pause/{workflow_id}")
def pause_workflow(workflow_id: str) -> Workflow:
    """Pause a specific workflow."""
    with app.state.state_handler.wc_state_lock():
        wf = app.state.state_handler.get_workflow(workflow_id)
        if wf.status in ["running", "in_progress", "queued"]:
            if wf.status == "running":
                send_admin_command_to_node("pause", wf.steps[wf.step_index].node)
                wf.steps[wf.step_index] = ActionStatus.PAUSED
            wf.paused = True
            app.state.state_handler.set_workflow(wf)

    return app.state.state_handler.get_workflow(workflow_id)


@app.get("/workflows/resume/{workflow_id}")
def resume_workflow(workflow_id: str) -> Workflow:
    """Resume a paused workflow."""
    with app.state.state_handler.wc_state_lock():
        wf = app.state.state_handler.get_workflow(workflow_id)
        if wf.paused:
            if wf.status == "running":
                send_admin_command_to_node("resume", wf.steps[wf.step_index].node)
                wf.steps[wf.step_index] = ActionStatus.RUNNING
            wf.paused = False
            app.state.state_handler.set_workflow(wf)
    return app.state.state_handler.get_workflow(workflow_id)


@app.get("/workflows/cancel/{workflow_id}")
def cancel_workflow(workflow_id: str) -> Workflow:
    """Cancel a specific workflow."""
    with app.state.state_handler.wc_state_lock():
        wf = app.state.state_handler.get_workflow(workflow_id)
        if wf.status == "running":
            send_admin_command_to_node("stop", wf.steps[wf.step_index].node)
            wf.steps[wf.step_index] = ActionStatus.CANCELLED
        wf.status = WorkflowStatus.CANCELLED
        app.state.state_handler.set_workflow(wf)
    return app.state.state_handler.get_workflow(workflow_id)


@app.get("/workflows/resubmit/{workflow_id}")
def resubmit_workflow(workflow_id: str) -> Workflow:
    """resubmit a previous workflow as a new workflow."""
    with app.state.state_handler.wc_state_lock():
        wf = app.state.state_handler.get_workflow(workflow_id)
        wf.workflow_id = new_ulid_str()
        wf.step_index = 0
        wf.start_time = None
        wf.end_time = None
        wf.submitted_time = datetime.now()
        for step in wf.steps:
            step.step_id = new_ulid_str()
            step.start_time = None
            step.end_time = None
            step.status = ActionStatus.NOT_STARTED
        copy_workflow_files(
            old_id=workflow_id,
            workflow=wf,
            working_directory=workcell.config.workcell_directory,
        )
        app.state.state_handler.set_workflow(wf)
    return app.state.state_handler.get_workflow(workflow_id)


@app.post("/workflows/retry")
def retry_workflow(workflow_id: str, index: int = -1) -> Workflow:
    """Retry an existing workflow from a specific step."""
    with app.state.state_handler.wc_state_lock():
        wf = app.state.state_handler.get_workflow(workflow_id)
        if wf.status in ["completed", "failed"]:
            if index >= 0:
                wf.step_index = index
            wf.status = WorkflowStatus.QUEUED
            app.state.state_handler.set_workflow(wf)
    return app.state.state_handler.get_workflow(workflow_id)


@app.post("/workflows/start")
async def start_workflow(
    workflow: Annotated[str, Form()],
    ownership_info: Annotated[Optional[str], Form()] = None,
    parameters: Annotated[Optional[str], Form()] = None,
    validate_only: Annotated[Optional[bool], Form()] = False,
    files: list[UploadFile] = [],
) -> Workflow:
    """
    parses the payload and workflow files, and then pushes a workflow job onto the redis queue

    Parameters
    ----------
    workflow: UploadFile
    - The workflow yaml file
    parameters: Optional[Dict[str, Any]] = {}
    - Dynamic values to insert into the workflow file
    ownership_info: Optional[OwnershipInfo]
    - Information about the experiments, users, etc. that own this workflow
    simulate: bool
    - whether to use real robots or not
    validate_only: bool
    - whether to validate the workflow without queueing it

    Returns
    -------
    response: Workflow
    - a workflow run object for the requested run_id
    """
    try:
        wf_def = WorkflowDefinition.model_validate_json(workflow)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=422, detail=str(e)) from e

    if ownership_info is None:
        ownership_info = OwnershipInfo()
    else:
        ownership_info = OwnershipInfo.model_validate_json(ownership_info)

    if parameters is None:
        parameters = {}
    else:
        parameters = json.loads(parameters)
        if not isinstance(parameters, dict) or not all(
            isinstance(k, str) for k in parameters
        ):
            raise HTTPException(
                status_code=400,
                detail="Parameters must be a dictionary with string keys",
            )
    workcell = app.state.state_handler.get_workcell()

    wf = create_workflow(
        workflow_def=wf_def,
        workcell=workcell,
        ownership_info=ownership_info,
        parameters=parameters,
        state_handler=app.state.state_handler,
    )

    if not validate_only:
        wf = save_workflow_files(
            working_directory=workcell.config.workcell_directory,
            workflow=wf,
            files=files,
        )
        with app.state.state_handler.wc_state_lock():
            app.state.state_handler.set_workflow(wf)
    return wf


# TODO add reserv nodes endpoint


if __name__ == "__main__":
    import uvicorn

    workcell = None
    workcell = WorkcellDefinition.load_model(require_unique=True)
    if workcell is None:
        raise ValueError(
            "No workcell manager definition found, please specify a path with --definition, or add it to your lab definition's 'managers' section"
        )
    uvicorn.run(
        app,
        host=workcell.config.host,
        port=workcell.config.port,
    )



"""Basic tests for workflow functionality"""

from pathlib import Path

from madsci.client.event_client import default_logger
from madsci.client.workcell_client import WorkcellClient

client = WorkcellClient("http://localhost:8013")

default_logger.log_info(client.get_node("liquid_handler"))
default_logger.log_info(
    client.add_node("liquid_handler", "http://localhost:2000", permanent=True)
)
wf = client.submit_workflow(
    Path("../../../../tests/example/workflows/test_workflow.workflow.yaml").resolve(),
    {},
)
default_logger.log_info(wf.workflow_id)
client.resubmit_workflow(wf.workflow_id)
default_logger.log_info(client.get_all_workflows())
